"""Ubiquiti mFi MPower exceptions"""
from __future__ import annotations


class MPowerError(Exception):
    """General mFi MPower error."""


class MPowerSSHError(MPowerError):
    """Error related to board info extraction via SSH."""


class MPowerSSHConnError(MPowerSSHError):
    """Error related to SSH connections."""


class MPowerSSHAuthError(MPowerSSHError):
    """Error related to SSH data authentication."""


class MPowerSSHReadError(MPowerSSHError):
    """Error related to SSH data reading."""


class MPowerSSHDataError(MPowerSSHError):
    """Error related to SSH data validity."""


class MPowerAPIError(MPowerError):
    """Error related to the "REST" API from Ubiquiti."""


class MPowerAPIConnError(MPowerAPIError):
    """Error related to "REST" API connections."""


class MPowerAPIAuthError(MPowerAPIError):
    """Error related to "REST" API authentication."""


class MPowerAPIReadError(MPowerAPIError):
    """Error related to "REST" API data reading."""


class MPowerAPIDataError(MPowerAPIError):
    """Error related to "REST" API data validity."""



"""Ubiquiti mFi MPower device"""
from __future__ import annotations

import json
from random import randrange
import ssl
import time

import aiohttp
import asyncio
from yarl import URL

from .board import MPowerBoard
from .entities import MPowerSensor, MPowerSwitch
from .exceptions import (
    MPowerAPIError,
    MPowerAPIConnError,
    MPowerAPIAuthError,
    MPowerAPIReadError,
    MPowerAPIDataError,
)


class MPowerDevice:
    """mFi mPower device representation."""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        use_ssl: bool = False,
        verify_ssl: bool = False,
        cache_time: float = 0.0,
        board_info: bool | None = None,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        """Initialize the device."""
        self.host = host
        self.url = URL(f"https://{host}" if use_ssl else f"http://{host}")
        self.username = username
        self.password = password
        self.cache_time = cache_time
        self._board_info = board_info

        self._board = MPowerBoard(self)

        if session is None:
            self._session_owned = True
            self._session = None
        else:
            self._session_owned = False
            self._session = session

        # NOTE: Ubiquiti mFi mPower Devices with firmware 2.1.11 use OpenSSL 1.0.0g (18 Jan 2012)
        if use_ssl:
            self._ssl = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
            self._ssl.set_ciphers("AES128-SHA:@SECLEVEL=0")
            self._ssl.verify_mode = ssl.CERT_REQUIRED if verify_ssl else ssl.CERT_NONE
            self._ssl.certs_loaded = False
        else:
            self._ssl = False

        self._cookie = (
            f"AIROS_SESSIONID={''.join([str(randrange(9)) for i in range(32)])}"
        )

        self._authenticated = False
        self._time = time.time()
        self._data = {}

    async def __aenter__(self) -> MPowerDevice:
        """Enter context manager scope."""
        await self.login()
        await self.update()
        return self

    async def __aexit__(self, *kwargs) -> None:
        """Leave context manager scope."""
        await self.logout()

    def __str__(self):
        """Represent this device as string."""
        if self._data:
            keys = ["name", "ipaddr", "hwaddr", "model"]
        else:
            keys = ["host"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    @property
    def name(self) -> str:
        """Return the device name."""
        if self.updated:
            try:
                return self.hostname
            except Exception:  # pylint: disable=broad-except
                pass
        return self.host

    @property
    def manufacturer(self) -> str:
        """Return the device manufacturer."""
        return "Ubiquiti"

    @property
    def board(self) -> MPowerBoard:
        """Return the device board."""
        return self._board

    @property
    def eu_model(self) -> bool | None:
        """Return whether this device is a EU model with type F sockets."""
        if self._board.updated:
            return self._board.eu_model
        return None

    async def request(
        self, method: str, url: str | URL, data: dict | None = None
    ) -> str:
        """Session wrapper for general requests."""
        url = URL(url)
        if not url.is_absolute():
            url = self.url / str(url).lstrip("/")
        try:
            async with self._session.request(
                method=method,
                url=url,
                headers={"Cookie": self._cookie},
                data=data,
                ssl=self._ssl,
                chunked=None,
            ) as resp:
                if resp.status != 200:
                    raise MPowerAPIReadError(
                        f"Received bad HTTP status code from device {self.name}: {resp.status}"
                    )

                # NOTE: Un-authorized request will redirect to /login.cgi
                if str(resp.url.path) == "/login.cgi":
                    self._authenticated = False
                else:
                    self._authenticated = True

                return await resp.text()
        except aiohttp.ClientSSLError as exc:
            raise MPowerAPIConnError(
                f"Could not verify SSL certificate of device {self.name}: {exc}"
            ) from exc
        except aiohttp.ClientError as exc:
            raise MPowerAPIConnError(
                f"Connection to device {self.name} failed: {exc}"
            ) from exc

    async def load_certs(self):
        if (self._ssl and self._ssl.verify_mode != ssl.CERT_NONE and not self._ssl.certs_loaded):
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, self._ssl.load_default_certs)
            self._ssl.certs_loaded = True

    async def login(self) -> None:
        """Login to this device."""
        if self._session_owned and self._session is None:
            self._session = aiohttp.ClientSession()
        if not self._authenticated:
            await self.load_certs()
            await self.request(
                "POST",
                "/login.cgi",
                data={"username": self.username, "password": self.password},
            )

            if not self._authenticated:
                raise MPowerAPIAuthError(
                    f"Login to device {self.name} failed due to wrong API credentials"
                )

    async def logout(self) -> None:
        """Logout from this device."""
        if self._authenticated:
            await self.request("POST", "/logout.cgi")
        if self._session_owned:
            await self._session.close()
            self._session = None

    async def update(self) -> None:
        """Update sensor data."""
        # NOTE: If board_info is
        #        - True, one attempt will be made (raise)
        #        - None, one attempt will be made (no raise)
        #        - False, no attempt will be made
        #       to update the board data!
        if not self._board.updated and self._board_info is not False:
            try:
                self._board = MPowerBoard(self)
                await self._board.update()
            except Exception as exc:  # pylint: disable=broad-except
                if self._board_info:
                    raise exc

        if not self._data or (time.time() - self._time) > self.cache_time:
            await self.login()
            text_status = await self.request("GET", "/status.cgi")
            text_sensors = await self.request("GET", "/mfi/sensors.cgi")

            try:
                data = json.loads(text_status)
                data.update(json.loads(text_sensors))
            except aiohttp.ContentTypeError as exc:
                raise MPowerAPIDataError(
                    f"Received invalid data from device {self.name}: {exc}"
                ) from exc

            status = data.get("status", None)
            if status != "success":
                raise MPowerAPIDataError(
                    f"Received invalid sensor update status from device {self.name}: {status}"
                )

            self._time = time.time()
            self._data = data

    @property
    def updated(self) -> bool:
        """Return if the device data has already been updated."""
        return bool(self._data)

    @property
    def data(self) -> dict:
        """Return device data."""
        if not self._data:
            raise MPowerAPIError(
                f"Device data for device {self.name} must be updated first"
            )
        return self._data

    @data.setter
    def data(self, data: dict) -> None:
        """Set device data."""
        self._data = data

    @property
    def host_data(self) -> dict:
        """Return the device host data."""
        return self.data.get("host", {})

    @property
    def fwversion(self) -> str:
        """Return the device host firmware version."""
        return self.host_data.get("fwversion", "")

    @property
    def hostname(self) -> str:
        """Return the device host name."""
        return self.host_data.get("hostname", "")

    @property
    def lan_data(self) -> dict:
        """Return the device LAN data."""
        return self.data.get("lan", {})

    @property
    def wlan_data(self) -> dict:
        """Return the device WLAN data."""
        return self.data.get("wlan", {})

    @property
    def ipaddr(self) -> str:
        """Return the device IP address from LAN if connected, else from WLAN."""
        lan_connected = self.lan_data.get("status", "") != "Unplugged"
        if lan_connected:
            return self.lan_data.get("ip", "")
        return self.wlan_data.get("ip", "")

    @property
    def hwaddr(self) -> str:
        """Return the device hardware address from LAN if connected, else from WLAN."""
        lan_connected = self.lan_data.get("status", "") != "Unplugged"
        if lan_connected:
            return self.lan_data.get("hwaddr", "")
        return self.wlan_data.get("hwaddr", "")

    @property
    def unique_id(self) -> str:
        """Return a unique device id from combined LAN/WLAN hardware addresses."""
        lan_hwaddr = self.lan_data.get("hwaddr", "")
        wlan_hwaddr = self.wlan_data.get("hwaddr", "")
        if lan_hwaddr and wlan_hwaddr:
            return f"{lan_hwaddr}-{wlan_hwaddr}"
        return ""

    @property
    def port_data(self) -> list[dict]:
        """Return the device port data."""
        return self.data.get("sensors", [])

    @property
    def ports(self) -> int:
        """Return the number of available device ports."""
        return len(self.port_data)

    @property
    def model(self) -> str:
        """Return the model name of this device as string."""
        if self._board.updated:
            return self._board.model
        ports = self.ports
        prefix = "mPower"
        suffix = " (EU)" if self.eu_model else ""
        if ports == 1:
            return f"{prefix} mini" + suffix
        if ports == 3:
            return prefix + suffix
        if ports in [6, 8]:
            return f"{prefix} Pro" + suffix
        return "Unknown"

    @property
    def description(self) -> str:
        """Return the device description as string."""
        ports = self.ports
        if ports == 1:
            return "mFi Power Adapter with Wi-Fi"
        if ports == 3:
            return "3-Port mFi Power Strip with Wi-Fi"
        if ports == 6:
            return "6-Port mFi Power Strip with Ethernet and Wi-Fi"
        if ports == 8:
            return "8-Port mFi Power Strip with Ethernet and Wi-Fi"
        return ""

    async def create_sensor(self, port: int) -> MPowerSensor:
        """Create a single sensor."""
        if not self.updated:
            await self.update()
        return MPowerSensor(self, port)

    async def create_sensors(self) -> list[MPowerSensor]:
        """Create all sensors as list."""
        if not self.updated:
            await self.update()
        return [MPowerSensor(self, i + 1) for i in range(self.ports)]

    async def create_switch(self, port: int) -> MPowerSwitch:
        """Create a single switch."""
        if not self.updated:
            await self.update()
        return MPowerSwitch(self, port)

    async def create_switches(self) -> list[MPowerSwitch]:
        """Create all switches as list."""
        if not self.updated:
            await self.update()
        return [MPowerSwitch(self, i + 1) for i in range(self.ports)]



"""Ubiquiti mFi MPower entities"""
from __future__ import annotations

from . import device  # pylint: disable=unused-import
from .exceptions import MPowerAPIDataError


class MPowerEntity:
    """mFi mPower entity representation."""

    def __init__(
        self,
        device: device.MPowerDevice,  # pylint: disable=redefined-outer-name
        port: int,
    ) -> None:
        """Initialize the entity."""
        self._device = device
        self._port = port

        if not device.updated:
            raise MPowerAPIDataError(f"Device {device.name} must be updated first")

        self._data = device.port_data[self._port - 1]

        if port < 1:
            raise ValueError(
                f"Port number {port} for device {device.name} is too small: 1-{device.ports}"
            )
        if port > device.ports:
            raise ValueError(
                f"Port number {port} for device {device.name} is too large: 1-{device.ports}"
            )

    def __str__(self):
        """Represent this entity as string."""
        name = f"name={self._device.name}"
        keys = ["port", "label"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({name}, {vals})"

    async def update(self) -> None:
        """Update entity data from device data."""
        await self._device.update()
        self._data = self._device.port_data[self._port - 1]

    @property
    def device(self) -> device.MPowerDevice:
        """Return the entity device."""
        return self._device

    @property
    def data(self) -> dict:
        """Return all entity data."""
        return self._data

    @data.setter
    def data(self, data: dict) -> None:
        """Set entity data."""
        self._data = data

    @property
    def unique_id(self) -> str:
        """Return unique entity id from unique device id and port."""
        return f"{self.device.unique_id}-{self.port}"

    @property
    def port(self) -> int:
        """Return the port number (starting with 1)."""
        return int(self._port)

    @property
    def label(self) -> str:
        """Return the entity label."""
        return str(self._data.get("label", ""))

    @property
    def output(self) -> bool:
        """Return the current output state."""
        return bool(self._data["output"])

    @property
    def relay(self) -> bool:
        """Return the initial output state which is applied after device boot."""
        return bool(self._data["relay"])

    @property
    def lock(self) -> bool:
        """Return the output lock state which prevents switching if enabled."""
        return bool(self._data["lock"])


class MPowerSensor(MPowerEntity):
    """mFi mPower sensor representation."""

    precision: dict[str, float | None] = {
        "power": None,
        "current": None,
        "voltage": None,
        "powerfactor": None,
    }

    def __str__(self):
        """Represent this sensor as string."""
        name = f"name={self._device.name}"
        keys = ["port", "label", "power", "current", "voltage", "powerfactor", "energy"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({name}, {vals})"

    def _value(self, key: str, scale: float = 1.0) -> float | None:
        """Process (scale and round) and return sensor value."""
        value = self._data.get(key)
        if value is not None:
            value *= scale
            precision = self.precision.get(key, None)
            if precision is not None:
                value = round(value, precision)
        return value

    @property
    def power(self) -> float | None:
        """Return the output power [W]."""
        return self._value("power")

    @property
    def current(self) -> float | None:
        """Return the output current [A]."""
        return self._value("current")

    @property
    def voltage(self) -> float | None:
        """Return the output voltage [V]."""
        return self._value("voltage")

    @property
    def powerfactor(self) -> float | None:
        """Return the output power factor ("real power" / "apparent power") [%]."""
        # NOTE: Raw data is dimensionless factor between 0 and 1
        return self._value("powerfactor", scale=100)

    @property
    def energy(self) -> float | None:
        """Return the energy (of this month) [kWh]."""
        # NOTE: Raw data is dimensionless impulse counter with 3200 imp/kWh
        return self._value("thismonth", scale=1/3200)


class MPowerSwitch(MPowerEntity):
    """mFi mPower switch representation."""

    def __str__(self):
        """Represent this switch as string."""
        name = f"name={self._device.name}"
        keys = ["port", "label", "output", "relay", "lock"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({name}, {vals})"

    async def set(self, output: bool, refresh: bool = True) -> None:
        """Set output to on/off."""
        await self._device.request(
            "POST", "/mfi/sensors.cgi", data={"id": self._port, "output": int(output)}
        )
        if refresh:
            await self.update()

    async def turn_on(self, refresh: bool = True) -> None:
        """Turn output on."""
        await self.set(True, refresh=refresh)

    async def turn_off(self, refresh: bool = True) -> None:
        """Turn output off."""
        await self.set(False, refresh=refresh)

    async def toggle(self, refresh: bool = True) -> None:
        """Toggle output."""
        await self.update()
        output = not bool(self._data["output"])
        await self.set(output, refresh=refresh)



"""This is the mFi mPower module."""
from __future__ import annotations

from .device import MPowerDevice
from .entities import MPowerSensor, MPowerSwitch
from .exceptions import *

__version__ = "1.3.5"



"""Ubiquiti mFi MPower board"""
from __future__ import annotations

import asyncssh

from . import device  # pylint: disable=unused-import
from .exceptions import MPowerAPIDataError
from .exceptions import (
    MPowerSSHConnError,
    MPowerSSHAuthError,
    MPowerSSHReadError,
    MPowerSSHDataError,
)


class MPowerBoard:
    """mFi mPower board representation."""

    # NOTE: Ubiquiti mFi mPower Devices with firmware 2.1.11 use Dropbear SSH 0.51 (27 Mar 2008).
    _ssh: dict = {
        "kex_algs": "diffie-hellman-group1-sha1",
        "encryption_algs": "aes128-cbc",
        # https://github.com/ronf/asyncssh/issues/263
        "server_host_key_algs": "ssh-rsa",
        # https://github.com/ronf/asyncssh/issues/132
        "known_hosts": None,
    }

    def __init__(
        self,
        device: device.MPowerDevice,  # pylint: disable=redefined-outer-name
    ) -> None:
        """Initialize the board."""
        self._device = device
        self._data = {}

    def __str__(self):
        """Represent this board as string."""
        if self._data:
            keys = [
                "name",
                "sysid",
                "cpurevision",
                "revision",
                "hwaddr",
                "eu_model",
                "model",
                "ports",
            ]
        else:
            keys = ["host"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    @property
    def host(self) -> str:
        """Return the board device host."""
        return self._device.host

    @property
    def name(self) -> str:
        """Return the board name."""
        try:
            return self.hostname
        except Exception:  # pylint: disable=broad-except
            return self.host

    async def update(self) -> None:
        """Update board data."""
        if not self._data:
            try:
                async with asyncssh.connect(
                    host=self.host,
                    username=self._device.username,
                    password=self._device.password,
                    **self._ssh,
                ) as conn:
                    # Read host name
                    result = await conn.run("cat /proc/sys/kernel/hostname")
                    status = result.exit_status
                    if status == 0:
                        try:
                            assert isinstance(result.stdout, str)
                            self._data["hostname"] = result.stdout.rstrip()
                        except Exception as exc:
                            raise MPowerSSHDataError(
                                f"Host name from device {self.host} is not valid: {exc}"
                            ) from exc
                    else:
                        raise MPowerSSHReadError(
                            f"Host name from device {self.host} could not be read: {status}"
                        )

                    # Read board data
                    result = await conn.run("cat /var/etc/board.info")
                    status = result.exit_status
                    if status == 0:
                        try:
                            assert isinstance(result.stdout, str)
                            for line in result.stdout.splitlines():
                                key_value = line.split("=")
                                if len(key_value) == 2:
                                    key, value = key_value
                                    self._data[key] = value
                        except Exception as exc:
                            raise MPowerSSHDataError(
                                f"Board data from device {self.host} is not valid: {exc}"
                            ) from exc
                    else:
                        raise MPowerSSHReadError(
                            f"Board data from device {self.host} could not be read: {status}"
                        )
            except asyncssh.PermissionDenied as exc:
                raise MPowerSSHAuthError(
                    f"Login to device {self.host} failed due to wrong SSH credentials"
                ) from exc
            except (OSError, asyncssh.Error) as exc:
                raise MPowerSSHConnError(
                    f"Connection to device {self.host} failed: {type(exc).__name__}({exc})"
                ) from exc

    @property
    def updated(self) -> bool:
        """Return if the board data has already been updated."""
        return bool(self._data)

    @property
    def data(self) -> dict:
        """Return board data."""
        if not self._data:
            raise MPowerAPIDataError(
                f"Board data for device {self.host} must be updated first"
            )
        return self._data

    @property
    def hostname(self) -> str:
        """Return if the board name."""
        return self.data.get("hostname", "")

    @property
    def sysid(self) -> str:
        """Return if the board system id."""
        return self.data.get("board.sysid", "")

    @property
    def cpurevision(self) -> str:
        """Return if the board CPU revision."""
        return self.data.get("board.cpurevision", "")

    @property
    def revision(self) -> str:
        """Return if the board revision."""
        return self.data.get("board.revision", "")

    @property
    def hwaddr(self) -> str:
        """Return if the board hardware address."""
        return self.data.get("board.hwaddr", "")

    @property
    def eu_model(self) -> bool | None:
        """Return whether this device is a EU model with type F sockets."""
        shortname = self.data.get("board.shortname", "")
        if len(shortname) > 2 and shortname.endswith("E"):
            return True
        elif len(shortname) > 1:
            return False
        else:
            return None

    @property
    def model(self) -> str:
        """Return if the board model."""
        name = self.data.get("board.name", "")
        eu_tag = " (EU)" if self.eu_model else ""
        if name:
            return name + eu_tag
        return ""

    @property
    def ports(self) -> int | None:
        """Return the number of available ports for the board device."""
        shortname = self.data.get("board.shortname", "")
        if len(shortname) > 1:
            return int(shortname[1])
        else:
            return None
